#include <stdio.h>
#include <string>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <signal.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <pwd.h>
#include <list>
#include <vector>
#include <map>
#include <algorithm>

#if __APPLE__
#define MSG_NOSIGNAL 0
typedef int socklen_t;
#endif

struct XoverRange
{
    int type;
    int low;
    int high;

    XoverRange() : type(0), low(0), high(0) {}
};

const int FROM_CACHE = 1;
const int FROM_SERVER = 2;

struct ConnectedSocket
{
    int s;
    std::string in;
    std::string out;
    bool can_write;

    ConnectedSocket() : s(-1), can_write(false) {}

    bool read();
    void write();
    void add_to_sets( int & n, fd_set * rp, fd_set * wp );
    bool is_set( fd_set * sp );
};

struct NNTPConnection
{
    ConnectedSocket     * server;
    ConnectedSocket     * client;
    std::string entering_group;
    std::string group;
    int server_state;
    int client_state;
    int xover_low;
    int xover_high;
    std::vector<XoverRange> xover_plan;
    int xover_plan_step;

    void check_server();
    void check_client();
    void start_xover( int low, int high );
    void next_xover_plan_step();

    NNTPConnection() : server(0), client(0), server_state(0), client_state(0), xover_plan_step(-1) {}
};
const int server_normal = 0;
const int server_cmd_xover = 1;
const int server_in_xover = 2;
const int server_cmd_group = 3;
const int server_cmd_data = 4;
const int server_in_data = 5;
const int server_cmd_post = 6;
const int server_cmd_xover_all = 7;
const int client_normal = 0;
const int client_data = 3;

typedef std::list<NNTPConnection *> NNTPConnection_l;
typedef std::list<NNTPConnection *>::iterator NNTPConnection_li;

std::string server_host = "";
int server_port = 119;
int listen_port = 0;
int listen_socket = -1;
std::string cache_dir = "";
std::string reader_cmd = "";
fd_set read_set;
fd_set write_set;
NNTPConnection_l connections;
bool running = true;
bool no_console = false;
bool debug_flag = false;
int trim_size = 0;

int usage( const char * prog );
bool make_listen_socket();
bool bind_listen_socket();
bool start_listening();
bool FFF( const char * msg );
void loop();
void new_client();
void read_console();
int connect_to_server();
std::string get_line( std::string & in );
std::vector<std::string> * split_into_words( std::string line );
XoverRange parse_xover( std::vector<std::string> * wp );
bool overview_in_cache( std::string group, int article_num );
std::string get_overview_from_cache( std::string group, int article_num );
void put_overview_in_cache( std::string group, int article_num, std::string data );
void start_reader( std::string cmd );
void reader_done( int sig );
void generic_sig( int sig ) { printf( "signal %d\n", sig ); }
void status( char c );
void trim_cache( int trim_size );
void trim_group( std::string group, int trim_size );

int main( int argc, char *argv[] )
{
    for ( int i = 1; i < argc; ++i )
    {
        std::string arg( argv[i] );
        if ( arg == "-d" )
            debug_flag = true;
        else if ( arg == "-p" )
        {
            if ( reader_cmd.length() || i+1 >= argc )
                return usage( argv[0] );
            listen_port = atoi( argv[i+1] );
            ++i;
        }
        else if ( arg == "-t" )
        {
            if ( reader_cmd.length() || i+1 >= argc )
                return usage( argv[0] );
            trim_size = atoi( argv[i+1] );
            ++i;
        }
        else if ( arg == "-s" )
        {
            if ( reader_cmd.length() || i+1 >= argc )
                return usage( argv[0] );
            server_host = std::string( argv[i+1] );
            ++i;
        }
        else if ( arg == "-c" )
        {
            if ( reader_cmd.length() || i+1 >= argc )
                return usage( argv[0] );
            cache_dir = argv[i+1];
            ++i;
        }
        else
        {
            if ( reader_cmd.length() )
                reader_cmd += ' ';
            reader_cmd += arg;
        }
    }

    if ( cache_dir.length() == 0 )
    {
        struct passwd * pp = getpwuid( geteuid() );
        if ( pp == NULL )
            return usage( argv[0] );
        cache_dir = std::string(pp->pw_dir) + "/.nntp_xover_cache";
    }

    mkdir( cache_dir.c_str(), 0700 );
    std::string real_cache = cache_dir + "/cache";
    mkdir( real_cache.c_str(), 0700 );
    if ( access( cache_dir.c_str(), F_OK ) == -1 || access( real_cache.c_str(), F_OK ) == -1 )
        return usage( argv[0] );

    if ( server_host.length() == 0 )
    {
        std::string config(cache_dir);
        config += "/server.txt";
        FILE * fp = fopen( config.c_str(), "rb" );
        if ( fp == NULL )
            return usage( argv[0] );
        char buf[512];
        int got;
        while ( (got = fread(buf, 1, sizeof buf, fp)) > 0 )
            server_host += std::string(buf,got);
        fclose( fp );
        while ( (got = server_host.length()) && server_host[got-1] < ' ' )
                server_host = std::string( server_host, 0, got-1 );
    }
    if ( server_host.length() == 0 )
        return usage( argv[0] );

    int pos = server_host.find( ':' );
    if ( pos != std::string::npos )
    {
        server_port = atoi( server_host.c_str() + pos + 1 );
        server_host = std::string( server_host, 0, pos );
    }

    if ( debug_flag && reader_cmd.length() )
    {
        fprintf( stderr, "Cannot use -d flag when specifying a reader command\n" );
        return 1;
    }
    if ( ! debug_flag && reader_cmd.length() == 0 )
    {
        std::string config(cache_dir);
        config += "/command.txt";
        FILE * fp = fopen( config.c_str(), "rb" );
        if ( fp )
        {
            char buf[512];
            int got;
            while ( (got = fread(buf, 1, sizeof buf, fp)) > 0 )
                reader_cmd += std::string(buf,got);
            fclose( fp );
            while ( (got = reader_cmd.length()) && reader_cmd[got-1] < ' ' )
                    reader_cmd = std::string( reader_cmd, 0, got-1 );
        }
    }

    if ( trim_size != 0 )
    {
        printf( "Triming cache down to %d items per newsgroup...\n", trim_size );
        trim_cache( trim_size );
    }

    printf( "server_host   [%s]\n", server_host.c_str() );
    printf( "server_port   [%d]\n", server_port );
    printf( "listen_port   [%d]\n", listen_port );
    printf( "cache_dir     [%s]\n", cache_dir.c_str() );
    printf( "reader_cmd    [%s]\n", reader_cmd.c_str() );

    if ( !make_listen_socket()
            || !bind_listen_socket()
            || !start_listening() )
        return 1;

    if ( reader_cmd.length() > 0 )
    {
        no_console = true;
        signal( SIGCHLD, reader_done );
        start_reader( reader_cmd );
        signal( SIGINT, SIG_IGN );
        signal( SIGQUIT, SIG_IGN );
    }
    loop();
    return 0;
}

int usage( const char * prog )
{
    fprintf( stderr, "Usage: %s [-d] [-t num] [-p port] [-s server[:port]] [-c cache_dir] [cmd [arg ...]]\n", prog );
    return 1;
}

bool make_listen_socket()
{
    listen_socket = socket( PF_INET, SOCK_STREAM, 0 );
    if ( listen_socket == -1 )
        return FFF("make_listen_socket: could not create socket");
    return true;
}

bool bind_listen_socket()
{
    struct sockaddr_in sin;
    sin.sin_family = PF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons( listen_port );
    int err = bind( listen_socket, (struct sockaddr *)&sin, sizeof sin );
    if ( err )
    {
        close( listen_socket );
        return FFF("bind_listen_socket: could not bind socket");
    }
    if ( listen_port == 0 )
    {   
        struct sockaddr_in sin;
        socklen_t len = sizeof sin;
        getsockname( listen_socket, (struct sockaddr *)&sin, &len );
        listen_port = ntohs( sin.sin_port );
        if ( ! no_console )
            printf( "Port=%d\n", listen_port );
    }
    return true;
}

bool start_listening()
{
    if ( listen( listen_socket, 5 ) != 0 )
        return FFF("start_listening: cannot listen");
    return true;
}

bool FFF( const char * msg )
{
    fprintf( stderr, "Error: %s\n", msg );
    return false;
}

void loop()
{
    while ( running ) {
        FD_ZERO( &read_set );
        FD_ZERO( &write_set );

        FD_SET( listen_socket, &read_set );
        int n = listen_socket + 1;

        if ( ! no_console )
            FD_SET( 0, &read_set );

        for ( NNTPConnection_li it = connections.begin(); it != connections.end(); ) {
            NNTPConnection * cp = *it;
            if ( cp->server->s == -1 && cp->client->s == -1 ) {
                NNTPConnection_li di = it++;
                connections.erase(di);
                delete cp->server;
                delete cp->client;
                delete cp;
                if ( ! no_console )
                    printf( "Deleted item from connection table\n" );
            } else {
                cp->server->add_to_sets( n, &read_set, cp->server->can_write ? 0 : &write_set );
                cp->client->add_to_sets( n, &read_set, cp->client->can_write ? 0 : &write_set );
                it++;
            }
        }

        struct timeval tv;
        tv.tv_sec = 1;
        tv.tv_usec = 0;
        n = select( n, &read_set, &write_set, 0, &tv );
        if ( n > 0 ) {
            if ( ! no_console && FD_ISSET( 0, &read_set ) )
                read_console();

            if ( FD_ISSET( listen_socket, &read_set ) )
                new_client();

            for ( NNTPConnection_li it = connections.begin(); it != connections.end(); ++it ) {
                NNTPConnection * cp = *it;

                if ( cp->server->is_set( &read_set ) )
                    if ( cp->server->read() )
                        cp->check_server();

                if ( cp->client->is_set( &read_set ) )
                    if ( cp->client->read() )
                        cp->check_client();

                if ( ! cp->client->can_write && cp->client->is_set( &write_set ) )
                    cp->client->write();

                if ( ! cp->server->can_write && cp->server->is_set( &write_set ) )
                    cp->server->write();
            }
        }
        if ( ! no_console )
            fflush( stdout );
    }
    close( listen_socket );
    for ( NNTPConnection_li it = connections.begin(); it != connections.end(); ++it ) {
        NNTPConnection * cp = *it;
        if ( cp->server->s != -1 ) {
            shutdown( cp->server->s, 2 );
            close( cp->server->s );
        }
        if ( cp->client->s != -1 ) {
            shutdown( cp->client->s, 2 );
            close( cp->client->s );
        }
    }
}

void read_console()
{
    char buf[2048];
    buf[0] = '\0';
    read( 0, buf, sizeof buf );
    if ( buf[0] == 'q' )
        running = false;
}

void new_client()
{
    ConnectedSocket * client = new ConnectedSocket;
    client->s = accept( listen_socket, 0, 0 );

    if ( ! no_console )
        printf( "accepted connection\n" );

    int ss = connect_to_server();
    if ( ss == -1 ) {
        close( client->s );
        delete client;
        return;
    }
    ConnectedSocket * server = new ConnectedSocket;
    server->s = ss;

    fcntl( client->s, F_SETFL, O_NONBLOCK );
    fcntl( server->s, F_SETFL, O_NONBLOCK );

    NNTPConnection * connection = new NNTPConnection;
    connection->server = server;
    connection->client = client;
    connections.push_back( connection );

    if ( ! no_console )
        printf( "connected to server\n" );
}

int connect_to_server()
{
    struct sockaddr_in name;
    memset( &name, 0, sizeof name );
    name.sin_family = PF_INET;
    name.sin_port = htons( server_port );
    name.sin_addr.s_addr = inet_addr( server_host.c_str() );
    if ( name.sin_addr.s_addr == INADDR_NONE )
    {
        struct hostent * hp = gethostbyname( server_host.c_str() );
        if ( hp )
            name.sin_addr.s_addr = *(long *)hp->h_addr_list[0];
    }
    if ( name.sin_addr.s_addr == INADDR_NONE )
        return -1;

    int s = socket( PF_INET, SOCK_STREAM, 0 );
    if ( s == -1 )
        return -1;

    int err = connect( s, (struct sockaddr *)&name, sizeof name );
    if ( err ) {
        close( s );
        s = -1;
    }
    return s;
}

std::string get_line( std::string & in )
{
    int pos = in.find( '\12' );
    if ( pos == std::string::npos )
        return "";
    std::string line( in, 0, pos+1 );
    in = std::string( in, pos+1, std::string::npos );
    return line;
}

std::vector<std::string> * split_into_words( std::string line )
{
    std::vector<std::string> * wp = new std::vector<std::string>;
    std::string word;
    int state = 0;
    for ( int i = 0; i < line.length(); ++i ) {
        char c = line[i];
        if ( state == 0 && c > ' ' ) {
                word += c;
                state = 1;
        }
        else if ( state == 1 ) {
            if ( c <= ' ' ) {
                wp->push_back( word );
                word = "";
                state = 0;
            } else
                word += c;
        }
    }
    return wp;
}

XoverRange parse_xover( std::vector<std::string> * wp )
{
    XoverRange r;
    if ( wp->size() < 2 )
        return r;
    r.low = atoi( (*wp)[1].c_str() );
    int pos = (*wp)[1].find( '-' );
    if ( pos == std::string::npos )
    {
        r.high = r.low;
        return r;
    }
    std::string highstr( (*wp)[1], pos+1, std::string::npos );
    r.high = atoi( highstr.c_str() );
    return r;
}

bool ConnectedSocket::read()
{
    if ( s == -1 )
        return false;
    char buf[2048];
    int got;
    bool got_some = false;
    while ( (got = recv(s, buf, sizeof buf, 0)) > 0 ) {
        in += std::string( buf, got );
        got_some = true;
    }
    if ( got == 0 ) {
        shutdown( s, 2 );
        close( s );
        s = -1;
        if ( ! no_console )
            printf( "EOF on recv...closing socket\n" );
        return true;
    }
    return got_some;
}

void ConnectedSocket::write()
{
    if ( out.length() && s != -1 )
    {
        can_write = true;
        int wrote = 0;
        while ( out.length() && (wrote = send(s, out.data(), out.length(), MSG_NOSIGNAL)) > 0 )
            out = std::string( out, wrote, std::string::npos );
        if ( out.length() > 0 )
            can_write = false;
    }
}

void ConnectedSocket::add_to_sets( int & n, fd_set * rp, fd_set * wp )
{
    if ( s != -1 && (rp || wp) ) {
        if ( rp )
            FD_SET( s, rp );
        if ( wp )
            FD_SET( s, wp );
        if ( s+1 > n )
            n = s+1;
    }
}

bool ConnectedSocket::is_set( fd_set * sp )
{
    if ( s != -1 )
        return FD_ISSET( s, sp ) != 0;
    else
        return false;
}

void NNTPConnection::check_server()
{
    std::string line;
    while ( (line = get_line( server->in )) != "" ) {
        switch ( server_state ) {
            case server_normal:
            {
                break;
            }
            case server_cmd_xover:
            {
                if ( line[0] == '2' ) {
                    server_state = server_in_xover;
                    if ( xover_plan.size() )
                        line = "";
                }
                break;
            }
            case server_in_xover:
            {
                if ( line[0] == '.' && line[1] < ' ' )
                {
                    server_state = server_normal;
                    for ( int i = xover_low; i <= xover_high; ++i )
                        if ( ! overview_in_cache( group, i ) )
                        {
                            status( 'M' );
                            put_overview_in_cache( group, i, std::string("") );
                        }
                    if ( xover_plan.size() )
                    {
                        line = "";
                        next_xover_plan_step();
                    }
                }
                else
                {
                    status( 'X' );
                    int article_num = atoi( line.c_str() );
                    put_overview_in_cache( group, article_num, line );
                }
                break;
            }
            case server_cmd_group:
            {
                server_state = server_normal;
                if ( line[0] == '2' ) {
                    group = entering_group;
                    entering_group = "";
                }
                break;
            }
            case server_cmd_xover_all:
            {
                server_state = server_normal;
                if ( line[0] == '2' ) {
                    std::vector<std::string> * wp = split_into_words( line );
                    int high_article = atoi( (*wp)[3].c_str() );
                    delete wp;
                    start_xover( xover_low, high_article );
                }
                break;
            }
            case server_cmd_data:
            {
                break;
            }
            case server_in_data:
            {
                break;
            }
            case server_cmd_post:
            {
                if ( line[0] == '2' )
                    client_state = client_data;
                else
                    status( 'p' );
                server_state = server_normal;
                break;
            }
        }
        client->out += line;
        client->write();
    }
    if ( server->s == -1 ) {
        if ( client->s != -1 ) {
            shutdown( client->s, 2 );
            close( client->s );
            client->s = -1;
            if ( ! no_console )
                printf( "server is closed...closed client\n" );
        }
    }
}

void NNTPConnection::check_client()
{
    std::string line;
    while ( (line = get_line( client->in )) != "" ) {
        std::vector<std::string> * wp = split_into_words( line );
        std::string cmd = wp->size() ? (*wp)[0] : "";
        switch ( client_state ) {
            case client_normal:
            {
                if ( strcasecmp( cmd.c_str(), "xover" ) == 0 )
                {
                    XoverRange r = parse_xover( wp );
                    if ( r.low != 0 )
                    {
                        if ( r.high != 0 )
                            start_xover( r.low, r.high );
                        else
                        {
                            xover_low = r.low;
                            server_state = server_cmd_xover_all;
                            server->out += "GROUP " + group + "\15\12";
                            server->write();
                        }
                        line = "";
                    }
                }
                else if ( strcasecmp( cmd.c_str(), "group" ) == 0 )
                {
                    status( 'G' );
                    server_state = server_cmd_group, entering_group = (*wp)[1];
                }
                else if ( strcasecmp( cmd.c_str(), "article" ) == 0 )
                {
                    status( 'A' );
                }
                else if ( strcasecmp( cmd.c_str(), "post" ) == 0 )
                {
                    server_state = server_cmd_post;
                }
                //else if ( strcasecmp( cmd.c_str(), "login" ) == 0 )
                //{
                //        server->out += "authinfo user XXXX\15\12";
                //        server->out += "authinfo pass XXXX\15\12";
                //        server->write();
                //        line = "";
                //}
                else
                {
                    status( '?' );
                    if ( ! no_console )
                        printf( "[%s]", cmd.c_str() );
                }
                break;
            }
            case client_data:
            {
                if ( cmd == "." )
                {
                    status( '*' );
                    client_state = client_normal;
                }
                break;
            }
        }
        delete wp;
        server->out += line;
        server->write();
    }
    if ( client->s == -1 ) {
        if ( server->s != -1 ) {
            shutdown( server->s, 2 );
            close( server->s );
            server->s = -1;
            if ( ! no_console )
                printf( "client is closed...closed server\n" );
        }
    }
}

void NNTPConnection::start_xover( int low, int high )
{
    xover_plan.resize(0);
    XoverRange r;
    r.low = low;
    r.high = low;
    r.type = (overview_in_cache( group, low ) ? FROM_CACHE : FROM_SERVER);
    for ( int i = low+1; i <= high; ++i )
    {
        int type = (overview_in_cache( group, i ) ? FROM_CACHE : FROM_SERVER);
        if ( r.type == type )
            r.high = i;
        else
        {
            xover_plan.push_back( r );
            r.low = i;
            r.high = i;
            r.type = type;
        }
    }
    xover_plan.push_back(r);
    client->out += "224 xover information follows\15\12";
    xover_plan_step = 0;
    next_xover_plan_step();
}

void NNTPConnection::next_xover_plan_step()
{
next_step:
    if ( xover_plan_step >= xover_plan.size() )
    {
        client->out += ".\15\12";
        client->write();
        xover_plan.resize(0);
        return;
    }

    XoverRange r = xover_plan[xover_plan_step++];
    if ( r.type == FROM_CACHE )
    {
        for ( int i = r.low; i <= r.high; ++i )
        {
            std::string data = get_overview_from_cache( group, i );
            client->out += data;
            if ( data.length() == 0 )
                status( '!' );
            else
                status( '.' );
        }
        goto next_step;
    }
    else
    {
        char cmd[64];
        sprintf( cmd, "XOVER %d-%d\15\12", r.low, r.high );
        if ( ! no_console )
            printf( "{%d+%d}", r.low, r.high-r.low+1 );
        server->out += cmd;
        server->write();
        server_state = server_cmd_xover;
        xover_low = r.low;
        xover_high = r.high;
    }
}

/////////////////////////////////////////////////////////
struct index_entry
{
    int article_num;
    int offset;
    int length;

    bool operator<( const index_entry & that ) const
    {
        return article_num < that.article_num;
    }
};

std::string open_group;                 // which group we have the data for
std::map<int,index_entry> article_to_index;

void load_group( std::string group );

bool overview_in_cache( std::string group, int article_num )
{
    if ( group != open_group )
        load_group( group );
    return (article_to_index.count(article_num) != 0);
}

std::string get_overview_from_cache( std::string group, int article_num )
{
    if ( group != open_group )
        load_group( group );
    index_entry e = article_to_index[article_num];
    if ( e.length == 0 )
        return "";
    std::string file = cache_dir + "/cache/" + group + ".overview";
    FILE * fp = fopen( file.c_str(), "rb" );
    if ( fp == NULL )
        return "";
    fseek( fp, e.offset, SEEK_SET );
    char * dp = new char[e.length];
    fread( dp, 1, e.length, fp );
    fclose( fp );
    std::string ret( dp, e.length );
    delete[] dp;
    return ret;
}

void put_overview_in_cache( std::string group, int article_num, std::string data )
{
    std::string ifile = cache_dir + "/cache/" + group + ".index";
    std::string ofile = cache_dir + "/cache/" + group + ".overview";

    if ( group != open_group )
        load_group( group );
    if ( article_to_index.count(article_num) != 0 )
    {
        index_entry e = article_to_index[article_num];
        if ( e.length != 0 )
            return;
    }
    index_entry e;
    e.article_num = article_num;
    if ( data.length() == 0 )
    {
        e.offset = 0;
        e.length = 0;
    }
    else
    {
        FILE * fp = fopen( ofile.c_str(), "ab" );
        e.offset = ftell( fp );
        e.length = data.length();
        fwrite( data.data(), 1, data.length(), fp );
        fclose( fp );
    }
    FILE * fp = fopen( ifile.c_str(), "ab" );
    fwrite( &e, sizeof e, 1, fp );
    fclose( fp );
    article_to_index[article_num] = e;
}

void load_group( std::string group )
{
    article_to_index.clear();
    open_group = group;

    std::string file = cache_dir + "/cache/" + group + ".index";
    FILE * fp = fopen( file.c_str(), "rb" );
    if ( ! fp )
    {
        fp = fopen( file.c_str(), "wb" );
        fclose( fp );
        file = cache_dir + "/cache/" + group + ".overview";
        fp = fopen( file.c_str(), "wb" );
        fclose( fp );
        return;
    }
    index_entry e;
    while ( fread( &e, sizeof e, 1, fp ) == 1 )
        article_to_index[e.article_num] = e;
    fclose( fp );
}

void start_reader( std::string cmd )
{
    if ( fork() != 0 )
        return;

    char aport[10];
    sprintf( aport, "%d", listen_port );

    close( listen_port );
    setenv( "NNTP_XOVER_CACHE", aport, 1 );
    execl( "/bin/sh", "/bin/sh", "-c", cmd.c_str(), 0 );
    printf( "ERROR: execl failed\n" );
    exit(0);
}

void reader_done( int sig )
{
    int status = 0;
    pid_t pid = waitpid( -1, &status, WNOHANG );
    if ( pid != 0 )
        running = false;
}

void status( char c )
{
    if ( ! no_console )
        putchar( c );
}

void trim_cache( int trim_size )
{
    std::vector<std::string> newsgroups;
    std::string real_cache = cache_dir + "/cache";
    DIR * dp = opendir( real_cache.c_str() );
    if ( dp == NULL )
    {
        fprintf( stderr, "Warning: could not open cache directory...groups not trimmed\n" );
        return;
    }
    struct dirent * ep;
    while ( (ep = readdir(dp)) != NULL )
    {
        std::string file(ep->d_name);
        int len = file.length();
        if ( len > 6 && std::string(file,len-6,6) == ".index" )
            newsgroups.push_back( std::string(file,0,len-6) );
    }
    closedir( dp );
    for ( int i = 0; i < newsgroups.size(); ++i )
        trim_group( newsgroups[i], trim_size );
}

void trim_group( std::string group, int trim_size )
{
    std::vector<index_entry> index;
    std::string index_file = cache_dir + "/cache/" + group + ".index";

    FILE * fp = fopen( index_file.c_str(), "rb" );
    if ( fp == NULL )
    {
        fprintf( stderr, "Warning: skipped group %s...could not open index file\n", group.c_str() );
        return;
    }
    index_entry e;
    while ( fread(&e, sizeof e, 1, fp) == 1 )
        index.push_back(e);
    fclose( fp );
    std::sort( index.rbegin(), index.rend() );
    if ( index.size() > trim_size )
    {
        int old_size = index.size();
        int new_size = 0;
        index.resize( trim_size );
        std::reverse( index.begin(), index.end() );
        for ( int i = 0; i < index.size(); ++i )
        {
            std::string file = cache_dir + "/cache/" + group + ".overview";
            FILE * fp = fopen( file.c_str(), "rb" );
            if ( fp == NULL )
            {
                fprintf( stderr, "Warning: skipped group %s...could not open overview file\n", group.c_str() );
                return;
            }
            fseek( fp, index[i].offset, SEEK_SET );
            char * dp = new char[index[i].length];
            fread( dp, 1, index[i].length, fp );
            fclose( fp );
            std::string data( dp, index[i].length );
            delete[] dp;
            put_overview_in_cache( "new."+group, index[i].article_num, data );
            if ( data.length() )
                ++new_size;
        }
        std::string dst = cache_dir + "/cache/" + group + ".index";
        std::string src = cache_dir + "/cache/new." + group + ".index";
        unlink( dst.c_str() );
        link( src.c_str(), dst.c_str() );
        unlink( src.c_str() );

        dst = cache_dir + "/cache/" + group + ".overview";
        src = cache_dir + "/cache/new." + group + ".overview";
        unlink( dst.c_str() );
        link( src.c_str(), dst.c_str() );
        unlink( src.c_str() );

        printf( "%s, %d --> %d\n", group.c_str(), old_size, new_size );
    }
}
